unit BDEDoRx1;
{
*********************************************************
*  demo for restructure / creation of local tables      *
*                                                       *
*  (c) 1996-97 Reinhard Kalinke                         *
*                                                       *
*********************************************************
}                 

{NOTE When compiling the samples or a project of your own using 
BDEDoRxS methods with Delphi 1 tests seem to indicate that you 
better increase stack size to 24 or even 32k.}

(*
 Demo of using DBIDoRestructure on the field definitions of
 Paradox and dBase tables from within Delphi v1.

 In the structure grid, Ctrl-Ins inserts a row, Ctrl-DEL deletes
 and Enter, Ctrl-Down or Ctrl-PgDn adds one. Moving fields is done
 by mouse drag and drop. There's not much validation coded here,
 you will have to make sure yourself to supply valid field data.

 The interface field data is taken from the DBD concept here but
 it should be easy to rewrite the FLDDescToList and ListToFLDDesc
 procs to fit any other terminology - if you have the source,
 that is...

 You may not use this code and the restructure unit BDEDoRxS
 unless otherwise granted by the author. Drop me a mail.

 (c) 1996-97,
 Reinhard Kalinke
 100417,3504 @compuserve.com

 NOTE The code was originally written to support TObjectList
 from TCONTAIN.PAS. Search the code for 'TDoRxList' if you want
 to replace it with your brand of TCollection object.
 (with the dcu-only version this cannot be changed...)

 -------------------------------------------------------

 NOTE In this demo implementation a simple 'fake' tcollection
 based on tlist is used.

 this is the interface of unit BDEDoRxS:

type
  EDoRxError = class(Exception);
  private
    FTableName: TFileName;
    procedure SetTableName(Value: TFileName);
  public
    property TableName: TFileName read FTableName write SetTableName;
  end;
  EDoRxProblem = class(EDoRxError);
  EDoRxKeyViol = class(EDoRxError);

 An exception of this type is raised when a problem table was
 create either due to data trimming with BDEFieldRestruct (will
 be EDoRxProblem then) or with records failing the RI constraint
 being set up with BDEAddRIContraint (EDoRxKeyViol). The property
 TableName will hold the name of that table for further processing.

 type {simple TList descendant that can free its objects}
  TDoRxList = class(TList)
  public
    procedure AddObject(Item: pointer);
    procedure FreeAll;
  end;

TDoRxList is used to hold the field definitions (details see
BDEFieldRestruct).

The procedures and functions:
-----------------------------

NOTE all of the functions below are designed and written for dBase
and Paradox tables only. So when a TableType input is required it
has to be ttDBASE or ttPARADOX.


Restructuring:
--------------

This one retrieves a table's physical field types:

procedure BDEGetFieldStructure(ATable: TTable;
                              AFieldDescList: TDoRxList);

This one does the restructuring. AskForTrim=true brings up a
dialog when data might be truncated, ShowProgrThreshold determines
the minimum recordcount that will bring up a progress form
(-1 means never show progress).
NOTE The callback for trimming data seems not be called on tables
with AutoInc fields.

procedure BDEFieldRestruct(ATable: TTable;
                          AFieldDescList: TDoRxList;
                          AskForTrim: boolean;
                          ShowProgrThreshold: longint);

The FieldDescList used in all above procs is a List of TStringList.
Each stringlist holds the info for one field in for following form
(just the same as it appears in the stringgrid of this demo):

stringlist[0] = fieldname,
stringlist[1] = type, using the single char abbreviations from DBD,
                e.g. 'F' is float, 'D' is date and so on
stringlist[2] = size, for dbase float and num seperated by a comma
stringlist[3] = field no. !!!MOST IMPORTANT!!!: in order to determine
                whether a field has been moved or modified, you will
                have to supply each field no in the *old* table when
                restructuring a table. Remember that you can change
                name, type, size and position in one single action
                - there has to be some sort of identifier. so for
                restructuring a table, first get the field data with
                BDEGetFieldStructure, modify it while keeping the
                field no, and feed the modified data back to
                BDEFieldRestruct.


procedure BDEOptRestruct(ATable: TTable; const OptNames,
                         OptValues: array of string);

BDEOptRestruct is bound for restructuring your tables with optional
parameters. You can thus change a Table's level, block size or
language driver etc. The number and order of OptNames and OptValues
have to match in order for the proc to work correctly. For OptNames
you can use literals or predefined constants. Here's a list of the
possible values and constants:

const
  {for use with both dBase and Paradox:}
  szLEVEL               = 'LEVEL';
  szLANGDRV             = 'LANGDRIVER';
  {for use with Paradox only:}
  szBLOCKSIZE           = 'BLOCK SIZE';
  szSTRICTRI            = 'STRICTINTEGRTY'; {note the spelling!}
  {for use with dBase only:}
  szMDXBLOCKSIZE        = 'MDX BLOCK SIZE';
  szMEMOBLOCKSIZE       = 'MEMO FILE BLOCK SIZE';

Note that for language drivers you have to supply the _short_ name of
the LangDrv you want to use (a list can be found in the BDE API help
file).

Here's an example of how to change the table level to 5 and language
driver to ansi intl on a Paradox table:

  BDEOptRestruct(Table1, [szLEVEL,szLANGDRV], ['5','ANSIINTL']);

Also have a look at the OptStrucBtnClick method of the sample project.


Creating tables:
----------------

The following two procs are for creating tables with optional
parameters:

procedure BDEPdoxCreate(ADataBase: TDataBase;
                       ATableName: TFileName;
                       AFieldDescList: TDoRxList;
                       ATableLevel, ABlockSize: smallint;
                       StrictRI: boolean);

{StrictRI is setting a flag bit when a table is created. However,
on getting a structure it seems it's not validated as long no refint
settings are actually applied. That's why the checkbox is set to
false when that tables structure is re-read after creation.}

procedure BDEdBaseCreate(ADataBase: TDataBase;
                        ATableName: TFileName;
                        AFieldDescList: TDoRxList;
                        ATableLevel, AMDXBlockSize,
                        AMemoBlockSize: smallint);

Both above procs use the following one for creating tables which
you might want to do as well. You have to additionally specify the
tabletype and you can supply any optional parameters in the same
way as described with BDEOptRestruct (see above). Advantages are
that you can for example define a language driver with table creation
or that you dont have to supply any values if not needed (just empty
brackets then...).

procedure BDETableCreate(ADataBase: TDataBase;
                        const ATableName: TFileName;
                        const ATableType: TTableType;
                        AFieldDescList: TDoRxList;
                        const OptNames,
                        OptValues: array of string);

'Borrow' structure
------------------

procedure BDECloneTable(ATable: TTable;
                       ADataBase: TDataBase; {}
                       const ANewTableName: TFileName;
                       const APassword: TCaption;
                       KeepRI: boolean);

This is used to create clones of existing tables. Supply a password
if you want to additionally encrypt the table or if the source table
is encrypted and you want to keep the aux passwords of the source
table. KeepRI determines whether the original RefInt checks are
copied as well (in most cases this not a good idea. Better
explicetely recreate them).


Referential Integrity
---------------------

The following functions can be used to maintain RI for your tables.

procedure BDEAddRIConstraint(ATable: TTable;
                            const OtherTable: TFileName;
                            const RIName: TCaption;
                            const RIModeOp, RIDelOp: RINTQual;
                            const StrictRI: boolean;
                            const ThisTableFields,
                            OtherTableFields: string);

where RINTQual = (rintRESTRICT, rintCASCADE).

Note: You have to add DBIProcs (D1) or BDE (D2) to your uses clause
to make use of these values.

I hope this bunch of parameters is kind of self-explaining. A few
notes though: this function is bound for the use on the dependent
table. The rintMASTER will be automatically added to the *.val file
of the master table. StrictRI affects *any* RI constraint on the
table, not just the current RI. As Paradox does not support cascaded
deletes yet, RIDelOp must be set to rintRESTRICT. I did include it
nevertheless for there is a chance that this will be added in later
BDE versions. ThisTableFields and OtherTableFields holds a list of
fieldnames separated by ';' (just like the Fields parameter of
TTable.AddIndex).

An example of using this function using DBDEMOS tables:

BDEAddRIConstraint(Table1, 'PARTS.DB', 'PartNoCheck',
                   rintCASCADE, rintRESTRICT,
                   True, 'PartNo', 'PartNo');

where Table1 is set to 'ITEMS.DB'. This call would add one of the
RI constraints the table is originally shipping with.


This is used to drop a RI constraint identified by it's name:

procedure BDEDropRIConstraint(ATable: TTable; const ARIName: string);


This one retrieves a list of existing RI's (by name) of the given
table:

procedure BDEGetRIList(ATable: TTable; AList: TStrings);

Don't miss to set up a TString*List* when not using this call with
a TListBox or the like...!

The procs calling the restructure/create methods are at the very
end of this unit. RI examples are in BDEDoRx2, the BDE* function
calls being at the very end again.

*)

{$B-}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, Grids, DBGrids, DB, DBTables, ExtCtrls,
  DBIProcs, DBITypes, DBIErrs, Spin, BDEDoRxS, Menus, Px7Table;

type
  TMainFrm = class(TForm)
    DataSource1: TDataSource;
    SelectButton: TButton;
    GetStrucBtn: TButton;
    RestructBtn: TButton;
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    Panel4: TPanel;
    DBGrid1: TDBGrid;
    StrucGrid: TStringGrid;
    OpenDialog1: TOpenDialog;
    Database1: TDatabase;
    CreateBtn: TButton;
    RadioGroup1: TRadioGroup;
    SpinEdit1: TSpinEdit;
    ClearBtn: TButton;
    CloseBtn: TButton;
    IndexBtn: TButton;
    RIBtn: TButton;
    SaveDialog1: TSaveDialog;
    SpinEdit2: TSpinEdit;
    Bevel2: TBevel;
    Label2: TLabel;
    Edit1: TEdit;
    Label1: TLabel;
    Bevel3: TBevel;
    Bevel4: TBevel;
    Label3: TLabel;
    SpinEdit3: TSpinEdit;
    Bevel5: TBevel;
    Bevel6: TBevel;
    Bevel1: TBevel;
    OptStrucBtn: TButton;
    PopupMenu1: TPopupMenu;
    SaveTableDefstofile1: TMenuItem;
    RestructureTablefromfile1: TMenuItem;
    Table1: TPx7Table;
    Bevel8: TBevel;
    N1: TMenuItem;
    Calculaterecordsize1: TMenuItem;
    PopupMenu2: TPopupMenu;
    PopupMenu3: TPopupMenu;
    Alpha1: TMenuItem;
    Numeric1: TMenuItem;
    SmallInt1: TMenuItem;
    Integer1: TMenuItem;
    Date1: TMenuItem;
    Time1: TMenuItem;
    TimeStamp1: TMenuItem;
    Logical1: TMenuItem;
    Memo1: TMenuItem;
    BLOB1: TMenuItem;
    OLEBLOB1: TMenuItem;
    Graphic1: TMenuItem;
    FormattedMemo1: TMenuItem;
    Bytes1: TMenuItem;
    BCD1: TMenuItem;
    AutoInc1: TMenuItem;
    Currency1: TMenuItem;
    Char1: TMenuItem;
    Numeric2: TMenuItem;
    Float1: TMenuItem;
    Bool1: TMenuItem;
    Date2: TMenuItem;
    Memo2: TMenuItem;
    BLOB2: TMenuItem;
    OLEBLOB2: TMenuItem;
    ProcessIndicesfromfile1: TMenuItem;
    EmptyTable1: TMenuItem;
    N2: TMenuItem;
    CopyTable1: TMenuItem;
    RenameTable1: TMenuItem;
    PackTable1: TMenuItem;
    ProtectTable1: TMenuItem;
    BDEversion1: TMenuItem;
    ComboBox1: TComboBox;
    Label4: TLabel;
    procedure Edit1Update(Sender: TObject);
    procedure SelectButtonClick(Sender: TObject);
    procedure GetStrucBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure RestructBtnClick(Sender: TObject);
    procedure StrucGridKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure StrucGridDragDrop(Sender, Source: TObject; X,
      Y: Integer);
    procedure StrucGridDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure StrucGridMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure StrucGridSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure StrucGridKeyPress(Sender: TObject; var Key: Char);
    procedure CreateBtnClick(Sender: TObject);
    procedure CloseBtnClick(Sender: TObject);
    procedure ClearBtnClick(Sender: TObject);
    procedure Table1AfterOpenClose(DataSet: TDataset);
    procedure RadioGroup1Click(Sender: TObject);
    procedure IndexBtnClick(Sender: TObject);
    procedure RIBtnClick(Sender: TObject);
    procedure OptStrucBtnClick(Sender: TObject);
    procedure PopupMenu1Popup(Sender: TObject);
    procedure SaveTableDefstofile1Click(Sender: TObject);
    procedure RestructureTablefromfile1Click(Sender: TObject);
    procedure Calculaterecordsize1Click(Sender: TObject);
    procedure Bytes1Click(Sender: TObject);
    procedure ProcessIndicesfromfile1Click(Sender: TObject);
    procedure EmptyTable1Click(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure CopyTable1Click(Sender: TObject);
    procedure RenameTable1Click(Sender: TObject);
    procedure ProtectTable1Click(Sender: TObject);
    procedure PackTable1Click(Sender: TObject);
    procedure BDEversion1Click(Sender: TObject);
  private
    { Private-Deklarationen }
    FFieldDescList: TDoRxList; {holds stringlists with field data}
    FWidth: integer;
    FHeight: integer;
    FPreventSizing: boolean;
    FInInternalOpen: boolean;
    procedure WMGetMinMaxInfo(var Msg: TWMGetMinMaxInfo);
              message WM_GETMINMAXINFO;
    procedure OpenTable(const AFileName: TFileName);
  public
    { Public-Deklarationen }
    FCurProps: CurProps; {cursor properties of current table}
    FIsPdx: boolean; {used for grid selectcell, set in SelectTable}
    procedure TryOpenExclusive(Tbl: TTable; bOpen, bExcl: boolean);
    function GetFileName(var NewName: TFileName): boolean;
  end;

var
  MainFrm: TMainFrm;

const
  PrimaryName = 'Primary *';
  NoIndexName = '- NONE -';

implementation

{$R *.DFM}

uses FileCtrl, Bdedorx2, Bdedorx3;

const CExt: array[boolean] of shortstring = ('db','dbf');
      CTblType: array[boolean] of shortstring = (szPARADOX,szDBASE);
      CGridChars: array[boolean] of array[boolean] of set of char
                  = ((['A','M','B','G','#','Y','F','O'],   {pdx size}
                      ['A','M','B','G','#','Y','F','O',     {pdx all}
                       'N','I','D','@','$','L','T','+','S']),
                     (['C','F','N'],          {dBase sizeable fields}
                      ['C','F','N','M','O','L','B','D']));{dBase all}

{--------}

function IsPdxTable(ATable: TTable): Boolean;
begin
  Result := (ATable.TableType = ttParadox) or
    (CompareText(ExtractFileExt(ATable.TableName),'.DB')=0);
end;

procedure SetControlState;
const CControlColor: array[0..1] of TColor
                     = (clBtnFace, clWindow);
      CSpinCaption:  array[0..1] of string
                     = ('Table Blocksize','MDX Blocksize');
      CSpinIncMin:   array[0..1] of smallint
                     = (1024, 512);
begin
  with MainFrm do
  begin
    if not Table1.Active then
      FIsPdx := not bool(RadioGroup1.ItemIndex);
    if FInInternalOpen then Exit;
    Edit1Update(Table1);
    CloseBtn.Enabled := Table1.Active;
    IndexBtn.Enabled := Table1.Active and Table1.Exclusive;
    GetStrucBtn.Enabled := Table1.Active;
    OptStrucBtn.Enabled := Table1.Active and Table1.Exclusive;
    RIBtn.Enabled := Table1.Active and FIsPdx;
    CreateBtn.Enabled := (StrucGrid.Rows[1][0] <> '');
    RestructBtn.Enabled := Table1.Active and Table1.Exclusive
                           and (StrucGrid.Rows[1][0] <> '');
    SpinEdit2.Increment := CSpinIncMin[RadioGroup1.ItemIndex];
    SpinEdit2.MinValue := CSpinIncMin[RadioGroup1.ItemIndex];
    Label2.Caption := CSpinCaption[RadioGroup1.ItemIndex];
    SpinEdit3.Color := CControlColor[RadioGroup1.ItemIndex]; {}
    SpinEdit3.Enabled := bool(RadioGroup1.ItemIndex);
    StrucGrid.PopUpMenu := TPopUpMenu(FindComponent('PopUpMenu'
                             +IntToStr(RadioGroup1.ItemIndex+2)));
    if Table1.Active then
      Caption := 'Restructure Demo - '+Table1.TableName
    else
      Caption := 'Restructure Demo';
    {CheckBox1.Enabled := not bool(RadioGroup1.ItemIndex); {}
  end;
end;

procedure InitGrid;
begin
  with MainFrm.StrucGrid do
  begin
    ColWidths[0] := (3*Canvas.TextWidth('Fieldname'));
    ColWidths[1] := Canvas.TextWidth(' Type ');
    ColWidths[2] := Canvas.TextWidth(' Type ');
    Rows[0].Add('Fieldname');
    Rows[0].Add('Type');
    Rows[0].Add('Size');
  end;
end;

procedure ClearGrid;
var i: smallint;
begin
  with MainFrm.StrucGrid do
  begin
    for i:=1 to pred(RowCount) do
      Rows[i].Clear;
    RowCount := 2;
  end;
  with MainFrm do
  begin
    if not Table1.Active then
      FIsPdx := not bool(RadioGroup1.ItemIndex);
    if FIsPdx then
    begin
      SpinEdit1.Value := StrToInt(szPDXLevel);
      SpinEdit2.Value := StrToInt(szPDXBlockSize);
      SpinEdit3.Value := 0;
      with ComboBox1 do
        ItemIndex := Items.IndexOf(szPDXLangDrv);
    end
    else
    begin
      SpinEdit1.Value := StrToInt(szDBFLevel);;
      SpinEdit2.Value := StrToInt(szDBFMDXBlockSize);
      SpinEdit3.Value := StrToInt(szDBFMemoBlockSize);
      with ComboBox1 do
        ItemIndex := Items.IndexOf(szDBFLangDrv);
    end;
  end;
end;

procedure PopulateGrid;
var i: smallint;
begin
  MainFrm.StrucGrid.OnSelectCell := nil;
  ClearGrid;
  with MainFrm.StrucGrid, MainFrm.FFieldDescList do
  begin
    RowCount := succ(Count);
    for i:= 0 to pred(Count) do
    begin
      Rows[succ(i)].Assign(TStringList(Items[i])); {tobjlist}
    end;
    OnSelectCell := MainFrm.StrucGridSelectCell;
    Col := 0;
  end;
end;

procedure GetDataFromGrid;
var i: smallint;
begin
  with MainFrm.FFieldDescList do
    begin
      FreeAll;
      for i:=1 to pred(MainFrm.StrucGrid.RowCount) do
      begin
        AddObject(TStringList.Create); {tobjlist}
        TStringList(Items[pred(i)]).AddStrings(MainFrm.StrucGrid.Rows[i]);
      end;
      Pack;
    end;
end;

{-- StringGrid tools --}

procedure AddGridRow(AGrid: TStringGrid);
begin
  with AGrid do
  begin
    RowCount := RowCount+1;
    Col := FixedCols;
    Row := pred(RowCount);
  end;
end;

procedure InsertGridRow(AGrid: TStringGrid; ANewRow: integer);
var i: integer;
begin
  with AGrid do
  begin
    RowCount := RowCount+1;
    for i := pred(RowCount) downto succ(ANewRow) do
    begin
      Rows[i].Assign(Rows[pred(i)]);
      Rows[pred(i)].Clear;
    end;
    Col := FixedCols;
  end;
end;

procedure DeleteGridRow(AGrid: TStringGrid; ADelRow: integer);
var i: smallint;
begin
  with AGrid do
  begin
    if (RowCount = 2) and (ADelRow = 1) then
    begin
      MessageBeep(0);
      Exit;
    end;
    Rows[ADelRow].Clear;
    for i := ADelRow to RowCount-1 do
    begin
      Rows[i].Assign(Rows[succ(i)]);
      Rows[succ(i)].Clear;
    end;
    RowCount := RowCount-1;
    if Row > pred(RowCount) then
      Row := pred(RowCount);
  end;
end;

procedure MoveGridRow(AGrid: TStringGrid; AMoveRow, ADestRow: integer);
var iShift: smallint;
begin
  with AGrid do
  begin
    {if (ADestRow = AMoveRow)
    or (DestRow > pred(RowCount)) then System.Exit; {}
    iShift := ord(ADestRow>AMoveRow);
    InsertGridRow(AGrid, ADestRow+iShift);
    inc(AMoveRow, abs(pred(iShift))); {ishift swapped<g>}
    Rows[ADestRow+iShift].Assign(Rows[AMoveRow]);
    DeleteGridRow(AGrid, AMoveRow);
    Row := ADestRow;
  end;
end;

{-- drag&drop within StrucGrid --}

procedure TMainFrm.StrucGridDragDrop(Sender, Source: TObject;
    X, Y: Integer);
var i: smallint;
    DCol, DRow: Longint;
begin
  with (Sender as TStringGrid) do
  begin
    MouseToCell( X, Y, DCol, DRow );
    if Row <> DRow then
      MoveGridRow((Sender as TStringGrid), Row, DRow);
  end;
end;

procedure TMainFrm.StrucGridDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
var DRow, DCol: longint;
begin
  with (Sender as TStringGrid) do
  begin
    MouseToCell( X, Y, DCol, DRow );
    Accept := (DRow <> 0);
    if Accept then
      DragCursor := crDrag
    else
      DragCursor := crNoDrop;
  end;
end;

procedure TMainFrm.StrucGridMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var DRow, DCol: longint;
begin
  if (Button = mbLeft) then
    with (Sender as TStringGrid) do
    begin
      MouseToCell( X, Y, DCol, DRow );
      if DRow <> 0 then
        BeginDrag(False);
    end;
end;

procedure TMainFrm.TryOpenExclusive(Tbl: TTable; bOpen, bExcl: boolean);
begin
  {make sure we have exclusive access/reset exclusive:}
  with Tbl do
  begin
    if (Exclusive<>bExcl) then
    begin
      if Active then Close;
      Exclusive := bExcl;
    end;
    try Active := bOpen;
    except {try again with exclusive = false:}
      if bExcl then Exclusive := False;
      Active := bOpen;
    end; {try}
  end; {with Tbl do}
end;

procedure TMainFrm.Edit1Update(Sender: TObject);
begin
  if Table1.Active then
  begin
    if not (Table1.IndexName = '') then
      Edit1.Text := Table1.IndexName
    else
      if StrPas(FCurProps.szTableType) = szPARADOX then
      begin
        Table1.IndexDefs.Update;
        if Table1.IndexDefs.Count > 0 then
          Edit1.Text := PrimaryName
        else
          Edit1.Text := ' not indexed';
      end
      else
        Edit1.Text := '';
  end
  else
    Edit1.Text := ' indexname';
end;

{-- StringGrid key handler --}

procedure TMainFrm.StrucGridKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Key = VK_RETURN) or (ssCtrl in Shift) then
    with StrucGrid do
    case Key of
      VK_INSERT: begin
                 InsertGridRow(StrucGrid, Row);
                 Rows[Row][3] := '0';
                 end;
      VK_RETURN,
      VK_DOWN,
      VK_NEXT:   begin
                 AddGridRow(StrucGrid);
                 Rows[Row][3] := '0';
                 end;
      VK_DELETE: DeleteGridRow(StrucGrid, Row);
    end;
  if (ssCtrl in Shift)
  and (Key in [VK_INSERT,VK_RETURN,VK_DOWN,VK_NEXT,VK_DELETE]) then
    Key := 0;
end;

{this can not really prevent invalid entries, however,...}
procedure TMainFrm.StrucGridSelectCell(Sender: TObject; Col, Row: Longint;
  var CanSelect: Boolean);
var x,y: array[0..1] of char;
begin
  if not Table1.Active then
    FIsPdx := not bool(RadioGroup1.ItemIndex);
  CanSelect := (Col = 0); {allow fieldname, deny fieldpos (col=4)}
  if (Col = 1) then
    CanSelect := (StrucGrid.Rows[Row][0] <> ''); {fieldtype}
  if (Col = 2) and (StrucGrid.Rows[Row][1] <> '') then
  begin
    StrPLCopy(x,StrucGrid.Rows[Row][1],1);
    if x[0] in   {size}
      CGridChars[not FIsPdx][False] then CanSelect := True
    else
      StrucGrid.Rows[Row][Col] := '';
  end;
  if not CreateBtn.Enabled and not RestructBtn.Enabled then
    SetControlState;
  if (not CanSelect) and (Col = 2) and (Row <> StrucGrid.Row) then
  begin
    StrucGrid.Col := 1;
    StrucGrid.Row := Row;
  end;
end;

{...nor can this}
procedure TMainFrm.StrucGridKeyPress(Sender: TObject; var Key: Char);
var CancelKey: boolean;
begin
  if (StrucGrid.Row = 1) and (StrucGrid.Rows[1][3] = '') then
    StrucGrid.Rows[1][3] := '0';
  if not Table1.Active then
    FIsPdx := not bool(RadioGroup1.ItemIndex);
  if Key in [#32..#255] then
    with (Sender as TStringGrid) do
    begin
      CancelKey := (Col<>0);
      if (Col = 1) then
        if (UpCase(Key) in CGridChars[not FIsPdx][True]) then
        begin
          CancelKey := False;
          if (length(Rows[Row][1]) >= 1) then
            Rows[Row][1] := '';
          if not (UpCase(Key) in CGridChars[not FIsPdx][False]) then
            Rows[Row][2] := '';
          Key := UpCase(Key);
        end;
      if (Col = 2) and (Key in ['0'..'9',',']) then
        CancelKey := False;
      if CancelKey then
      begin
        MessageBeep(0);
        Key := #0;
      end;
    end;
end;


{-- Buttons etc --}

 {without a TDatabase component pointing to the current
 database dir, some DBI functions that use DBIFormFullName
 and thus need a valid DBHandle (hDBIdb) will most probably
 fail. Seems like an implicit VCL database is set to the
 startup dir once and never updated. so if you don't like
 'invalid handle to the function' messages, always use a
 TDatabase and avoid assigning full path tablenames.}

procedure AssignDBDir(ADataBase: TDataBase; const AFileName: TFileName);
begin
  with ADataBase do
  if (Params.Count = 0)
  or (Params[0] <> 'PATH='+ExtractFilePath(AFileName)) then
  begin
    if Connected then Connected := False;
    DriverName := 'STANDARD'; {clears any alias as well}
    Params.Clear;
    Params.Add('PATH='+ExtractFilePath(AFileName));
    Open;
  end;
end;

procedure TMainFrm.OpenTable(const AFileName: TFileName);
begin
  with Table1, RadioGroup1 do
  begin
    ClearGrid;
    AssignDBDir(DataBase1, AFileName);
    Table1.Close;
    Table1.IndexName := '';
    Table1.TableName := ExtractFileName(AFileName);
    Table1.Open;
    TryOpenExclusive(Table1, True, True); {}
    FIsPdx := IsPdxTable(Table1);
    ItemIndex := ord(not FIsPdx);
    GetStrucBtnClick(Table1);
  end;
end;

procedure TMainFrm.SelectButtonClick(Sender: TObject);
begin
  if OpenDialog1.Execute then
  begin
    OpenTable(OpenDialog1.FileName);
  end;
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.GetStrucBtnClick(Sender: TObject);
begin
  BDEGetFieldStructure(Table1, FFieldDescList);
  PopulateGrid;
  with ComboBox1 do
    ItemIndex := Items.IndexOf(StrPas(FCurProps.szLangDriver));
  SpinEdit1.Value := FCurProps.iTblLevel;
  if FIsPdx then
  begin
    SpinEdit2.Value := FCurProps.iBlockSize*1024;
  end
  else
  begin
    FInInternalOpen := True;
    SpinEdit2.Value := BDEGetMDXBlockSize(Table1);
    SpinEdit3.Value := BDEGetMemoBlockSize(Table1);
    FInInternalOpen := False;
  end;
  SetControlState;
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.CloseBtnClick(Sender: TObject);
begin
  Table1.Close;
  Application.ProcessMessages;
end;

procedure TMainFrm.ClearBtnClick(Sender: TObject);
begin
  ClearGrid;
  SetControlState;
end;

procedure TMainFrm.RadioGroup1Click(Sender: TObject);
begin
  with RadioGroup1, Table1 do
    if Active then
      Active := (ItemIndex = ord(not FIsPdx))
    else
      SetControlState;
  ClearGrid;
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.Table1AfterOpenClose(DataSet: TDataset);
begin
  if Dataset.Active then
    DBIGetCursorProps((Dataset as TDBDataset).Handle, FCurProps)
  else
  begin
    FillChar(FCurProps,sizeOf(CURProps),#0);
    if IdxInspFrm <> nil then
    begin
      IdxInspFrm.Close;
    end;
  end;
  SetControlState;
end;

function GetEnvStr(VarName : string): string;
var
  Len     : Word;
  EnvStz  : PChar;
  NameStz : array[ 0..180 ] of Char;
begin
  StrPCopy( NameStz, VarName );   { Convert VarName to PChar }
  Len := StrLen( NameStz );

  {$IFDEF WIN32}
  EnvStz := GetEnvironmentStrings; { EnvStz holds entire env }
  {$ELSE}
  EnvStz := GetDosEnvironment;     { EnvStz holds entire env }
  {$ENDIF}

  while EnvStz^ <> #0 do
  begin                 { Pick off Variable Name and Compare }
    if (StrLIComp(EnvStz, NameStz, Len ) = 0 )
    and (EnvStz[Len] = '=' ) then
    begin        { Convert to Pascal string before returning }
      Result := StrPas(EnvStz + Len + 1);
      Exit;
    end;
    Inc(EnvStz, StrLen(EnvStz) + 1);      { Jump to Next Var }
  end;
  Result := '';
end;
{ This code came from Lloyd's help file! }
{ (with just a little modification to make it fit for WIN32) }

procedure TMainFrm.FormCreate(Sender: TObject);
var TempDir: string;
begin
  {creating FieldDescList:}
  FFieldDescList := TDoRxList.Create;
  {initializing controls:}
  InitGrid;
  ClearGrid;
  BDEGetLangDrivers(ComboBox1.Items);
  with ComboBox1 do
    ItemIndex := Items.IndexOf(szPDXLangDrv);
  SetControlState;
  {setting up Session.PrivateDir to the temp dir:
  (NOTE this is just q&d and does only work when a
  valid temp dir has been set up with your system)}
  TempDir := GetEnvStr('TEMP');
  if (TempDir > '') and DirectoryExists(TempDir) then
    Session.PrivateDir := TempDir;
  {NOTE you should set up a previnst test to avoid
   multiple instances of your app. There are samples
   and components in the GO DELPHI/BDELPHI libs.}
  BDEWatchdBaseIndices;
end;

procedure TMainFrm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Table1.Close;
  Application.ProcessMessages;
  FFieldDescList.FreeAll;
  FFieldDescList.Free;
end;


{------ calling the restructure unit --------}

procedure TMainFrm.RestructBtnClick(Sender: TObject);
begin
  Screen.Cursor := crHourGlass;
  try
    GetDataFromGrid;
    try
      BDEFieldRestruct(Table1, FFieldDescList, True, 1000);
    except
      on E:EDoRxProblem do
        ShowMessage(E.Message)
      else raise;
    end;
  finally
    Screen.Cursor := crDefault;
    GetStrucBtnClick(Sender);
  end;
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.CreateBtnClick(Sender: TObject);
begin
  GetDataFromGrid;
  with SaveDialog1 do
  begin
    DefaultExt := CExt[not FIsPdx];
    FilterIndex := succ(ord(not FIsPdx));
    if Execute and not (ofExtensionDifferent in Options) then
    begin
      AssignDBDir(DataBase1, FileName);
      case RadioGroup1.ItemIndex of
       0: BDEPdoxCreate(DataBase1, FileName,
                        FFieldDescList,
                        SpinEdit1.Value,   {tbllevel}
                        SpinEdit2.Value,   {block size}
                        False,             {strictRI}
                        ComboBox1.Text);   {LangDriver}
       1: BDEdBaseCreate(DataBase1, FileName,
                        FFieldDescList,
                        SpinEdit1.Value,   {tbllevel}
                        SpinEdit2.Value,   {mdx block size}
                        SpinEdit3.Value,   {memo block size}
                        ComboBox1.Text);   {LangDriver}
      end;
      OpenTable(FileName);
    end;
  end;
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.IndexBtnClick(Sender: TObject);
begin
  if (IdxInspFrm <> nil) then
    IdxInspFrm.BringToFront
  else
  begin
    IdxInspFrm := TIdxInspFrm.Create(Application);
    try
      IdxInspFrm.ShowModal;
    except
      TryOpenExclusive(Table1, True, True);
      raise;
    end;
  end;
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.RIBtnClick(Sender: TObject);
begin
  with TRefIntFrm.Create(Application) do
  try
    ShowModal;
    {if ModalResult = mrOk then
      GetStrucBtnClick(Sender); {not appreciated...}
  finally
    Free;
  end;
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.OptStrucBtnClick(Sender: TObject);
const CBoolVerb: array[boolean] of shortstring = ('FALSE','TRUE');
begin
  if FIsPdx then
    {call is using literals}
    BDEOptRestruct(Table1, ['LEVEL','BLOCK SIZE','LANGDRIVER'],
                   [IntToStr(SpinEdit1.Value),IntToStr(SpinEdit2.Value),
                    ComboBox1.Text])
  else
    {call is using the constants}
    BDEOptRestruct(Table1, [szLEVEL,szMDXBLOCKSIZE,szMEMOBLOCKSIZE,szLANGDRV],
                   [IntToStr(SpinEdit1.Value),IntToStr(SpinEdit2.Value),
                    IntToStr(SpinEdit3.Value),ComboBox1.Text]);
  GetStrucBtnClick(Sender);
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.PopupMenu1Popup(Sender: TObject);
begin
  SaveTableDefstofile1.Enabled := Table1.Active;
  RestructureTablefromfile1.Enabled := Table1.Active;
  ProcessIndicesfromfile1.Enabled := Table1.Active;
  EmptyTable1.Enabled := Table1.Active;
  Calculaterecordsize1.Enabled := (FIsPdx
                                    or (RadioGroup1.ItemIndex = 0))
                                  and (StrucGrid.Cells[1,1] <> '');
end;

procedure TMainFrm.SaveTableDefstofile1Click(Sender: TObject);
begin
  with SaveDialog1 do
  begin
    FilterIndex := 5;
    FileName := ChangeFileExt(Table1.TableName,'.dbi');
    if Execute then
      BDESaveTableDefsToFile(Table1,ChangeFileExt(FileName,'.dbi'));
    FilterIndex := 3;
  end;
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.RestructureTablefromfile1Click(Sender: TObject);
begin
  with OpenDialog1 do
  try
    FilterIndex := 5;
    FileName := ChangeFileExt(Table1.TableName,'.dbi');
    if Execute then
    begin
      Screen.Cursor := crHourGlass;
      BDERestructTableFromFile(Table1, FileName);
    end;
  finally
    Screen.Cursor := crDefault;
    GetStrucBtnClick(Sender);
    FilterIndex := 3;
  end;
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.ProcessIndicesfromfile1Click(Sender: TObject);
begin
  with OpenDialog1 do
  try
    FilterIndex := 5;
    FileName := ChangeFileExt(Table1.TableName,'.dbi');
    if Execute then
    begin
      Screen.Cursor := crHourGlass;
      BDEAddIndicesFromFile(Table1, FileName);
    end;
  finally
    Screen.Cursor := crDefault;
    GetStrucBtnClick(Sender);
    FilterIndex := 3;
  end;
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.Calculaterecordsize1Click(Sender: TObject);
var MaxRecs, RecSize, RecsPerBlock, FreeSpace: longint;
    i: integer;
begin
  RecSize := 0;
  with StrucGrid do
  begin
    for i:=0 to pred(RowCount) do
    begin
      case Cells[1,i][1] of
       'A': RecSize := RecSize + StrToInt(Cells[2,i]);
       'D','T','I','+': RecSize := RecSize + 4;
       'N','$','Y','@': RecSize := RecSize + 8;
       'M','B','F','O','G': RecSize := RecSize + 10 + StrToInt(Cells[2,i]);
       'S': RecSize := RecSize + 2;
       'L': RecSize := RecSize + 1;
      end;
    end;
  end;
  RecsPerBlock := (SpinEdit2.Value-6) div RecSize;
  FreeSpace := (SpinEdit2.Value-6)-(RecSize*RecsPerBlock);
  MaxRecs := 65536*RecsPerBlock;
  ShowMessage('Record Size is: '+IntToStr(RecSize)+' bytes'+#13#10
             +'Records per Block: '+IntToStr(RecsPerBlock)+#13#10
             +'Unused Space per Block: '+IntToStr(FreeSpace)+' bytes'+#13#10
             +'Max No of Records in Table: '+FormatFloat('###############,',MaxRecs));
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.Bytes1Click(Sender: TObject);
begin
  with StrucGrid do
    Cells[1,Row] := (Sender as TMenuItem).Caption[1];
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.WMGetMinMaxInfo(var Msg: TWMGetMinMaxInfo);
begin
  inherited;
  if FPreventSizing then
    with (self), Msg.MinMaxInfo^ do
    begin
      ptMinTrackSize.x:= FWidth;
      ptMinTrackSize.y:= FHeight;
    end;
end;

procedure TMainFrm.FormShow(Sender: TObject);
begin
  FWidth := Width;
  FHeight := Height;
  FPreventSizing := True;
end;

function TMainFrm.GetFileName(var NewName: TFileName): boolean;
begin
  Result := False;
  with SaveDialog1 do
  begin
    DefaultExt := copy(ExtractFileExt(Table1.TableName),2,3);
    if (Uppercase(DefaultExt) = 'DBF') then
      Filter := 'dBase tables|*.dbf'
    else
      Filter := 'Paradox tables|*.db';
    Result := Execute;
    if Result then
      NewName := Filename;
    if Result and (ofExtensionDifferent in Options) then
    begin
      Result := not Result;
      {raising an invalid filename error:}
      DBIError(DBIERR_INVALIDFILENAME);
    end;
  end; {with SaveDialog1}
end;


procedure TMainFrm.CopyTable1Click(Sender: TObject);
var NewName: TFileName;
begin
  if GetFileName(NewName) then
    BDECopyTable(Table1,NewName,True);
end;

procedure TMainFrm.RenameTable1Click(Sender: TObject);
var NewName: TFileName;
begin
  if GetFileName(NewName) then
    BDERenameTable(Table1,NewName);
end;

procedure TMainFrm.ProtectTable1Click(Sender: TObject);
var Password: string;
    PromptAdd: string;
begin
  PromptAdd := '';
  Password := '';
  {if there is a pw on the table:}
  if FCurProps.bProtected then
    PromptAdd := ' / Ok to remove current';
  if InputQuery('Password',
                'Enter new password'+PromptAdd,
                 Password) then
  begin
    Session.AddPassword(Password);
    BDEProtectTable(Table1,Password);
  end;
end;

procedure TMainFrm.PackTable1Click(Sender: TObject);
begin
  BDEPackTable(Table1);
end;

procedure TMainFrm.EmptyTable1Click(Sender: TObject);
begin
  if MessageDlg('You are about do delete the contents of this table.'+#13#10#13#10
               +'Proceed?',mtWarning,[mbYes,mbNo],0) = mrYes then
  begin
    BDEEmptyTable(Table1);
  end;
  ActiveControl := StrucGrid;
end;

procedure TMainFrm.BDEversion1Click(Sender: TObject);
begin
  {$IFNDEF WIN32}
  ShowMessage('Current BDE version is: v'+BDEGetIdapi16Version);
  {$ELSE}
  ShowMessage('Current BDE version is: v'+BDEGetIdapi32Version);
  {$ENDIF}
end;

end.             {======= end of code ========}




