unit BDEDoRx3;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls, DB, DBIProcs, DBITypes, DBIErrs,
  Tabs, DBTables;

type
  TIdxInspFrm = class(TForm)
    Panel1: TPanel;
    Memo1: TMemo;
    Panel2: TPanel;
    Panel3: TPanel;
    AddIndexBtn: TButton;
    DeleteIndexBtn: TButton;
    ReIndexAllBtn: TButton;
    AlterIndexBtn: TButton;
    CloseBtn: TButton;
    Panel4: TPanel;
    TabSet1: TTabSet;
    Notebook1: TNotebook;
    ListBox1: TListBox;
    ApplyIndexBtn: TButton;
    ExpandBtn: TButton;
    ListBox2: TListBox;
    procedure ListBox1Click(Sender: TObject);
    procedure ListBox2Click(Sender: TObject);
    procedure ReIndexAllBtnClick(Sender: TObject);
    procedure AlterIndexBtnClick(Sender: TObject);
    procedure AddIndexBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure TabSet1Click(Sender: TObject);
    procedure DeleteIndexBtnClick(Sender: TObject);
    procedure ExpandBtnClick(Sender: TObject);
    procedure ApplyIndexBtnClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure CloseBtnClick(Sender: TObject);
  private
    { Private-Deklarationen }
    FIndexDesc: IDXDesc;
    FPreventSizing: boolean;
    procedure GetIndices(Sender: TObject);
    procedure GetIndexTypes(Sender: TObject);
    procedure SetButtonStatus(Sender: TObject);
    procedure WMGetMinMaxInfo(var Msg: TWMGetMinMaxInfo);
              message WM_GETMINMAXINFO;
    procedure WMInitMenuPopup(var Msg: TWMInitMenuPopup);
              message WM_INITMENUPOPUP;
    procedure WMNCHitTest(var Msg: TWMNCHitTest);
              message WM_NCHitTest;
  public
    { Public-Deklarationen }
  end;

var
  IdxInspFrm: TIdxInspFrm;

implementation

uses BDEDoRx1, BDEDoRx4, BDEDoRxS;

{$R *.DFM}
{$B-;XL+}

const sbVerb: array[boolean] of string = ('False','True');


procedure TIdxInspFrm.ApplyIndexBtnClick(Sender: TObject);
begin
  with MainFrm.Table1, ListBox1 do
  begin
    if (Items[ItemIndex] = PrimaryName)
    or (Items[ItemIndex] = NoIndexName) then
      IndexName := ''
    else
      IndexName := Items[ItemIndex];
  MainFrm.Edit1Update(Sender);
  SetButtonStatus(Sender);
  MainFrm.Table1.Refresh; {}
  {MainFrm.Table1.Close;
  MainFrm.Table1.Open; {}
  end;
end;

procedure TIdxInspFrm.AddIndexBtnClick(Sender: TObject);
begin
  with TAddIdxFrm.Create(Application) do
  try
    ShowModal;
    if (ModalResult = mrOk) then
      GetIndices(Sender);
  finally
    Free;
  end;
end;

procedure TIdxInspFrm.DeleteIndexBtnClick(Sender: TObject);
begin
 try
  with ListBox1 do
    if (Items[ItemIndex] = PrimaryName) then
    begin
      if (MessageDlg('Deleting the primary index means deleting all indices!'+#13#10
                   +'Continue anyway?',mtConfirmation,[mbYes,mbCancel],0) = mrYes) then
      begin
        MainFrm.Table1.IndexName := '';
        BDEDropAllIndices(MainFrm.Table1);
      end;
    end
    else
      if (ItemIndex <> -1) then
        MainFrm.Table1.DeleteIndex(Items[ItemIndex]);
 finally
  GetIndices(Sender);
 end;
end;

procedure TIdxInspFrm.ReIndexAllBtnClick(Sender: TObject);
begin
  with MainFrm do
    Check( DBIRegenIndexes(Table1.Handle) );
end;

procedure TIdxInspFrm.AlterIndexBtnClick(Sender: TObject);
begin
  with TAddIdxFrm.Create(Application) do
  try
    Tag := 1;
    FIndexName := ListBox1.Items[ListBox1.ItemIndex];
    if (FIndexName = PrimaryName) then
      FIndexName := '';
    ShowModal;
    if (ModalResult = mrOk) then
      GetIndices(Sender);
  finally
    Free;
  end;
end;

procedure TIdxInspFrm.SetButtonStatus(Sender: TObject);
begin
  with ListBox1, NoteBook1, MainFrm.Table1 do
  begin
    ApplyIndexBtn.Enabled := (ItemIndex <> -1)
                             and (PageIndex = 0);
    ReindexAllBtn.Enabled := (Items.Count <> 0)
                             and (PageIndex = 0)
                             and (IndexName = '')
                             and (Exclusive);
    AlterIndexBtn.Enabled := (ItemIndex <> -1)
                             and (PageIndex = 0)
                             and (Items[ItemIndex]
                                  <>NoIndexName)
                             and (IndexName = '')
                             and (Exclusive);
    DeleteIndexBtn.Enabled := (ItemIndex <> -1)
                              and (PageIndex = 0)
                              {and (Items[ItemIndex]
                                   <>PrimaryName) {}
                              and (Items[ItemIndex]
                                   <>NoIndexName)
                              and (IndexName = '')
                              and (Exclusive);
    AddIndexBtn.Enabled := (IndexName = '');
  end;
end;

procedure TIdxInspFrm.GetIndices(Sender: TObject);
begin
  if (Sender = AddIndexBtn) then
  begin
    DBIGetCursorProps(MainFrm.Table1.Handle,MainFrm.FCurProps);
    MainFrm.Edit1Update(Sender);
  end;
  with Memo1, ListBox1 do
  begin
    Lines.Clear;
    Items.Clear;
    MainFrm.Table1.IndexDefs.Update;
    MainFrm.Table1.GetIndexNames(Items);
    if (StrPas(MainFrm.FCurProps.szTableType) = szPARADOX)
    and (MainFrm.FCurProps.iIndexes > Items.Count) then
      Items.Insert(0,PrimaryName)
    else
      if (Items.Count = 0) then
        Items.Add(NoIndexName);
    if (Items.Count > 0) then
      ItemIndex := 0
    else
      ItemIndex := -1;
  end;
  TabSet1.TabIndex := 0;
  TabSet1Click(Sender);
  SetButtonStatus(Sender);
end;

procedure TIdxInspFrm.FormCreate(Sender: TObject);
begin
  TabSet1.Tabs := NoteBook1.Pages;
  NoteBook1.PageIndex := TabSet1.TabIndex;
  GetIndexTypes(Sender);
  GetIndices(Sender);
  ExpandBtnClick(Sender);
  CalcControlSize(self);
end;

procedure TIdxInspFrm.TabSet1Click(Sender: TObject);
begin
  if TabSet1.TabIndex = 0 then
    ListBox1Click(Sender)
  else
    ListBox2Click(Sender);
  NoteBook1.PageIndex := TabSet1.TabIndex;
  SetButtonStatus(Sender);
end;

procedure TIdxInspFrm.GetIndexTypes(Sender: TObject);
var hIdxCur: hDBICur;
    pIdxBuf: pByte;
    IdxProps: CURProps;
begin
  pIdxBuf := nil;
  ListBox1.Items.Clear;
  try
    Check(DBIOpenIndexTypesList(MainFrm.FCurProps.szTableType, hIdxCur));
    Check(DBIGetCursorProps(hIdxCur,IdxProps));
    GetMem(pIdxBuf,IdxProps.iRecSize);
    while not (DBIGetNextRecord(hIdxCur,dbiNOLOCK,pIdxBuf,nil)
               = DBIERR_EOF) do
      ListBox2.Items.Add(pIDXType(pIdxBuf)^.szName);
  finally
    DBICloseCursor(hIdxCur);
    if pIdxBuf<>nil then
      FreeMem(pIdxBuf,IdxProps.iRecSize);
  end;
  TabSet1.Enabled := True;
  if (ListBox2.Items.Count > 0) then
    ListBox2.ItemIndex := 0
  else
    ListBox2.ItemIndex := -1;
end;

procedure TIdxInspFrm.ExpandBtnClick(Sender: TObject);
const ExpandCaps: array[boolean] of string = (' +',' -');
var i: integer;
begin
  with Panel1 do
  begin
    if not Visible then
      Panel2.Align := alLeft;
    Visible := not Visible;
    Memo1.Visible := Visible;
    if Visible then
    begin
      Height := Panel2.Height;
      Memo1.Align := alClient;
    end;
    FPreventSizing := False;
    CalcControlSize(self);
    if not Visible then
      Panel2.Align := alClient;
    FPreventSizing := not Visible;
    ExpandBtn.Caption := ExpandCaps[Visible];
  end;
end;

procedure TIdxInspFrm.ListBox1Click(Sender: TObject);
var BDERes: DBIResult;
    sKeyFields, sDescFields: string;
    sIndexName: string;
    i: integer;
begin
  with Memo1.Lines, FIndexDesc do
  begin
    if (ListBox1.ItemIndex = -1) then System.Exit;
    BeginUpdate;
    Clear;
    Add('TableName:    '+#9+ExtractFileName(MainFrm.Table1.TableName));
    Add('TableLevel:   '+#9+IntToStr(MainFrm.FCurProps.iTblLevel));
    Add('LangDriver:   '+#9+StrPas(MainFrm.FCurProps.szLangDriver));
    Add('');
    try
      if (ListBox1.Items[ListBox1.ItemIndex] = NoIndexName) then
        System.Exit;
      if (ListBox1.Items[ListBox1.ItemIndex] = PrimaryName) then
        sIndexName := ''
      else
        sIndexName := ListBox1.Items[ListBox1.ItemIndex];
      for i:=1 to MainFrm.FCurProps.iIndexes do
      begin
        Check(DBIGetIndexDesc(MainFrm.Table1.Handle,
                              i,FIndexDesc));
        if (CompareText(StrPas(szName),sIndexName) = 0) then
          Break;
      end;

      sKeyFields := #9;
      sDescFields := #9;
      Add('IndexName:    '+#9+ExtractFileName(StrPas(szName)));
      if (StrPas(MainFrm.FCurProps.szTableType) = szPARADOX) then
        Add('IndexID:      '+#9+IntToStr(iIndexId))
      else
        Add('IndexTagName: '+#9+StrPas(szTagName));
      Add('Format:       '+#9+StrPas(szFormat));
      Add('KeyLen:       '+#9+IntToStr(iKeyLen));
      Add('BlockSize:    '+#9+IntToStr(iBlockSize));
      Add('RestructNo:   '+#9+IntToStr(iRestrNum));
      Add('Primary:      '+#9+sbVerb[bPrimary]);
      Add('Unique:       '+#9+sbVerb[bUnique]);
      Add('Descending:   '+#9+sbVerb[bDescending]);
      Add('Maintained:   '+#9+sbVerb[bMaintained]);
      Add('CaseInSens.:  '+#9+sbVerb[bCaseInsensitive]);
      Add('OutOfDate:    '+#9+sbVerb[bOutOfDate]);
      Add('FieldsInKey:  '+#9+IntToStr(iFldsInKey));
      for i:=0 to pred(iFldsInKey) do
        sKeyFields := sKeyFields+IntToStr(aiKeyFld[i])+',';
      {$IFDEF WIN32}
      SetLength(SKeyFields,pred(Length(SKeyFields)));
      {$ELSE}
      sKeyFields[0] := pred(sKeyFields[0]);{}
      {$ENDIF}
      Add('KeyField(s):   '+sKeyFields);
      if bDescending
      and (StrPas(MainFrm.FCurProps.szTableType) = szPARADOX) then
      begin
        {'1' indicates that the corresponding field in the
        aiKeyFld array is sorted descending:}
        {$IFDEF WIN32}
        for i:=0 to pred(iFldsInKey) do
          sDescFields := sDescFields+IntToStr(ord(abDescending[i]))+',';
        SetLength(SDescFields,pred(Length(SDescFields)));
        {$ELSE}
        for i:=0 to pred(iFldsInKey) do
          sDescFields := sDescFields+IntToStr(iUnUsed[i])+',';
        sDescFields[0] := pred(sDescFields[0]);{}
        {$ENDIF}
        Add('DescField(s):  '+sDescFields);
      end;
      if (StrPas(MainFrm.FCurProps.szTableType) = szDBASE) then
      begin
        Add('-- dBase only --');
        Add('ExprIndex:    '+#9+sbVerb[bExpIdx]);
        if bExpIdx then
        begin
          Add('Expression: ');
          Add('  '+StrPas(szKeyExp));
        end;
        Add('SubSet:       '+#9+sbVerb[bSubset]);
        if bSubset then
        begin
          Add('Condition: ');
          Add('  '+StrPas(szKeyCond));
        end;
      end;
    finally
      EndUpdate;
      SetButtonStatus(Sender);
    end;
  end;
end;

procedure TIdxInspFrm.ListBox2Click(Sender: TObject);
var BDERes: DBIResult;
    IndexTypeDesc: IDXType;
    pIndexTypeName: DBIName;
    KeyFieldNos: string;
    i: integer;
begin
  if (ListBox2.ItemIndex = -1) then Exit;
  StrPCopy(pIndexTypeName, ListBox2.Items[ListBox2.ItemIndex]);
  with Memo1.Lines, IndexTypeDesc do
  begin
    BeginUpdate;
    Clear;
    Add('DriverName:    '+#9+StrPas(MainFrm.FCurProps.szTableType));
    try

      Check( DBIGetIndexTypeDesc(MainFrm.FCurProps.szTableType,
                                    pIndexTypeName,
                                    IndexTypeDesc) );

      Add('Index ID:      '+#9+IntToStr(iId));
      Add('IndexTypeName: '+#9+StrPas(szName));
      Add('Description:');
      Add('  '+StrPas(szText));
      Add('Format:        '+#9+StrPas(szFormat));
      Add('Composite:     '+#9+sbVerb[bComposite]);
      Add('Primary:       '+#9+sbVerb[bPrimary]);
      Add('Unique:        '+#9+sbVerb[bUnique]);
      Add('KeyDescending: '+#9+sbVerb[bKeyDescending]);
      Add('FldDescending: '+#9+sbVerb[bFldDescending]);
      Add('Maintained:    '+#9+sbVerb[bMaintained]);
      Add('KeyExpression: '+#9+sbVerb[bKeyExpr]);
      Add('Subset:        '+#9+sbVerb[bSubset]);
      Add('CaseInSens.:   '+#9+sbVerb[bCaseInsensitive]);
      Add('---------------------------');
      Add('Supported');
      Add('  by Delphi:   '+#9+sbVerb[bMaintained]);
    finally
      EndUpdate;
    end;
  end;
end;


procedure TIdxInspFrm.WMGetMinMaxInfo(var Msg: TWMGetMinMaxInfo);
begin
  inherited;
  if FPreventSizing then
    with (self), Msg.MinMaxInfo^ do
    begin
      ptMinTrackSize.x:= Width;
      ptMaxTrackSize.x:= Width;
      ptMinTrackSize.y:= Panel3.Height+2*(Panel3.Height div 3);
      ptMaxTrackSize.y:= Screen.Height-Top-12;
    end;
end;

procedure TIdxInspFrm.WMInitMenuPopup(var Msg: TWMInitMenuPopup);
begin
  inherited;
  if FPreventSizing and Msg.SystemMenu then
    EnableMenuItem(Msg.MenuPopup, SC_SIZE, MF_BYCOMMAND or MF_GRAYED)
end;

procedure TIdxInspFrm.WMNCHitTest(var Msg: TWMNCHitTest);
begin
  inherited;
  if FPreventSizing then
    with Msg do
      if Result in [HTLEFT, HTRIGHT, HTBOTTOMRIGHT,
                    HTBOTTOMLEFT, HTTOPRIGHT, HTTOPLEFT] then
         Result:= HTNOWHERE;
end;



procedure TIdxInspFrm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Action := caFree;
  IdxInspFrm := nil;
end;

procedure TIdxInspFrm.CloseBtnClick(Sender: TObject);
begin
  (self).Close;
end;

initialization
begin
  IdxInspFrm := nil;
end;

end.
