program Hash1;

{$APPTYPE CONSOLE}

uses
  SysUtils;

const
  ElementCount = 53;

type
  THashFunction = function (const S : string) : integer;

function CalcHash1(const S : string) : integer;
begin
  Result := length(S) mod ElementCount;
end;

function CalcHash2(const S : string) : integer;
begin
  Result := ord(S[1]) mod ElementCount;
end;

function CalcHash3(const S : string) : integer;
var
  i : integer;
begin
  Result := 0;
  for i := 1 to length(S) do
    inc(Result, ord(S[i]));
  Result := Result mod ElementCount;
end;

function CalcELFHash(const S : string) : integer;
var
  G : longint;
  i : integer;
begin
  Result := 0;
  for i := 1 to length(S) do begin
    Result := (Result shl 4) + ord(S[i]);
    G := Result and $F0000000;
    if (G <> 0) then
      Result := Result xor (G shr 24);
    Result := Result and (not G);
  end;
  Result := Result mod ElementCount;
end;

function CalcRandomHash(const S : string) : integer;
begin
  Result := Random(ElementCount);
end;

const
  TPEmployees : array [0..32] of string[15] =
    ('Antypas',          'Bolduc',           'Brady',
     'Bucknall',         'Cappellucci',      'Cope',
     'DelRossi',         'Douglas',          'Fairweather',
     'Foley',            'Frerking',         'Geiger',
     'Ghinaudo',         'Horaz',            'Huffman',
     'Hughes',           'Inman',            'Kokkonen',
     'Larsen',           'Leier',            'Medlin',
     'Milner',           'Phillips',         'Reisdorph',
     'Roberts',          'Rogers',           'Rose',
     'Salisbury',        'Trickey',          'Troxell',
     'Turner',           'Warner',           'Welch');

procedure RunHashTest(HashFunc : THashFunction; HashNum : integer);
var
  i, j, Count : integer;
  Hashes : array [0..ElementCount-1] of integer;
begin
  writeln('---Hash ', HashNum, '---');
  for i := 0 to ElementCount-1 do
    Hashes[i] := 0;
  for i := 0 to 32 do
    inc(Hashes[HashFunc(TPEmployees[i])]);
  for j := 0 to 15 do begin
    Count := 0;
    for i := 0 to ElementCount-1 do
      if (Hashes[i] = j) then
        inc(Count);
    write(j:6, ':' , Count:3);
  end;
  writeln;
  readln;
end;

function IntPower(x : double; y : integer) : double;
var
  i : integer;
begin
  Result := 1.0;
  for i := 1 to y do
    Result := Result * x;
end;

function Factorial(y : integer) : double;
var
  i : integer;
begin
  if (y <= 1) then
    Result := 1.0
  else begin
    Result := 1.0;
    for i := 1 to y do
      Result := Result * i;
  end;
end;

procedure DisplayExpectedDistribution;
var
  RdivN         : double;
  EtominusRdivN : double;
  i, Count      : integer;
begin
  writeln('---Expected---');
  RdivN := (High(TPEmployees) + 1.0) / ElementCount;
  EtominusRdivN := exp(-RdivN);
  for i := 0 to 15 do begin
    Count := Round(ElementCount *
                   IntPower(RdivN, i) *
                   EtominusRdivN /
                   Factorial(i));
    write(i:6, ':' , Count:3);
  end;
  writeln;
  readln;
end;

begin
  Randomize;
  RunHashTest(CalcHash1, 1);
  RunHashTest(CalcHash2, 2);
  RunHashTest(CalcHash3, 3);
  RunHashTest(CalcELFHash, 4);
  RunHashTest(CalcRandomHash, 5);
  DisplayExpectedDistribution;
end.
